/* Most applications that control USBIO directly need to import
   xyzzy.hardware.UsbIo class only.  */
import xyzzy.hardware.UsbIo;

/** A demonstration to flush an LED on USBIO.  The LED is assumed to
    be wired to P1.0, i.e., the bit 0 (LSB) of Port #1 (high current
    port.)  */
class Led
{
    /** The time in milliseconds the flushing continues.  */
    static final int duration = 5000;	// 5 seconds.

    /** LED flushing interval in milliseconds.  */
    static final int interval = 200;	// 0.2 seconds.

    /** The main entry to the demo application.  We ignore arguments,
        BTW.  */
    public static void main ( String[] argv )
    {
	/* Just catch any Exception/Error here, and report them.  This
           is just for a demonstration/debugging purposes.  */
	try {

	    /* Pick up a USBIO device.  (One is arbitrarily chosen, if
               two or more are attached.  Remember, this is just a
               demonstration.)  */
	    final UsbIo.Device usbio = UsbIo.getDevice();
	    if ( usbio == null ) {
		System.err.println( "No USBIO is attached." );
		return;
	    }

	    /* We have an (opened) USBIO device object.  It must be
               closed when we finished using it.  To ensure it, it is
               recommended to use try-finally blocks.  */
	    try {

		/* bits is an internal copy of the Port #1 state.
                   USBIO initializes it to 0xFF, so we start with the
                   same value.  */
		int bits = 0xFF;

		/* We want to */
		for ( int count = duration / interval; count > 0; --count ) {

		    /* Reverse the (internal) lit state of the LED and
                       reflect it to the extenal (real) LED (with
                       WRITE PORT 1 command.)  */
		    bits ^= 0x01;
		    usbio.invoke( UsbIo.WRITE1, bits );

		    /* Stop execution for the specified interval.  */
		    Thread.sleep( interval );

		}

	    } finally {

		/* We have to close the USBIO device object before
                   releasing it.  Also, it is usually better to put
                   all USBIO pins in their initial states, before
                   exiting.  */
		usbio.invoke( UsbIo.WRITE0, 0xFF );
		usbio.invoke( UsbIo.WRITE1, 0xFF );
		usbio.close();
	    }

	} catch ( Throwable e ) {
	    /* Show any Error/Exception.  */
	    e.printStackTrace();
	}
    }
}
