import java.text.DecimalFormat;
import xyzzy.hardware.UsbIo;

/** A small application to measure the communication speed (i.e.,
    command/response delay) of Morphy USBIO.  It shows two separate
    results: one for synchronous (i.e., the completion of the command
    is verified by waiting for the response) and another for
    asynchronous.  Since Morphy USBIO is a low-speed HID device, and
    the response is carried by a (low-speed) interrupt pipe, the
    theoretical limit (calculated from USB spec.) would be 10ms and
    1ms respectively.  This program also shows an auxiliarly
    iformation, which roughly shows the effective _polling_ rate for
    the interrupt pipe.  Test the actual performance by yourself.  It
    should take about 10 to 20 seconds to run.  */
class Speed
{
    /** Number of samples to try, i.e., same command will be issued
        <code>samples</code> times, and the entire elapsed time will
        be measured.  */
    static final int samples = 1000;

    /** The main entry point.  */
    public static void main ( final String[] argv )
    {
	try {
	    final long t0 = System.currentTimeMillis();

	    /* Get the USBIO device object to test.  */
	    final UsbIo.Device usbio = UsbIo.getDevice();
	    if ( usbio == null ) {
		System.out.println( "No Morphy USBIO found." );
		return;
	    }

	    try {
		System.out.println( "Testing..." );

		/* Step 1: measure synchronous speed.  */
		usbio.invoke( UsbIo.READ0, 0 );
		final long t11 = System.currentTimeMillis();
		for ( int i = 0; i < samples; i++ ) {
		    usbio.invoke( UsbIo.READ0, 0 );
		}
		final long t12 = System.currentTimeMillis();
		showResult( "synchronous delay", t11, t12 );

		/* Step 2: measure asynchronous speed.  */
		usbio.invoke( UsbIo.WRITE0 | UsbIo.ASYNC, 0xFF );
		final long t21 = System.currentTimeMillis();
		for ( int i = 0; i < samples; i++ ) {
		    usbio.invoke( UsbIo.WRITE0 | UsbIo.ASYNC, 0xFF );
		}
		final long t22 = System.currentTimeMillis();
		showResult( "asynchronous delay", t21, t22 );

		/* Step 3: measure interrupt pipe polling rate.  */
		usbio.invoke( UsbIo.RECEIVE, 0 );
		final long t31 = System.currentTimeMillis();
		for ( int i = 0; i < samples; i++ ) {
		    usbio.invoke( UsbIo.RECEIVE, 0 );
		}
		final long t32 = System.currentTimeMillis();
		showResult( "interrupt polling", t31, t32 );

	    } finally {
		usbio.close();
	    }

	    final long t4 = System.currentTimeMillis();
	    System.out.println( "total testing time: " + ( t4 - t0 ) + "ms" );
	} catch ( Throwable e ) {
	    e.printStackTrace();
	}
    }

    static void showResult ( final String type, final long t1, final long t2 )
    {
	final DecimalFormat df = new DecimalFormat( "0.000" );
	final double delay = ( double )( t2 - t1 ) / samples;
	System.out.println( type + " "
			    + df.format( delay ) + "ms"
			    + " (" + df.format( 1 / delay ) + "KB/s)" );
    }
}


