import java.awt.*;
import java.awt.event.*;
import xyzzy.hardware.UsbIo;
import xyzzy.hardware.UsbIoSpiDevice;
import xyzzy.hardware.Ds1620;
import xyzzy.awt.ThermometerGraph;

/** A sample application for USBIO/DS1620 USB Digital Thermometer.  */
class Thermo
{
    /** The entry point as a standalone application.<br>
	
	Parameters to the application are as follows:<br>

	-v: Specify the visual (Window-based) mode operation.<br>
	-t: Specify the text (console) mode operation.<br>
	-c: DS1620 configuration mode.<br>

	Options are exclusive, i.e., at most one option is allowed.
	If no option is specified, the operation defaults to -v (vsual
	mode.)<br>

	When -v option is specified (or no option is specified), the
	application opens an awt window and shows a Thermometer style
	bar graph.  The bar graph will be updated periodically (in 5
	to 10 seconds interval,) showing the temperature.  The program
	runs until the window is closed.<br>

	When -t option is specified, the application shows the current
	temperature on the System.out.  As a demonstration, two
	temperature values are shown; one is the value measured by
	Ds1620.getTemperature(), and another by
	Ds1620.getHiResTenperature().  The program stops after showing
	temperatures.<br>

	The -c option may take an optional parameter.  If no parameter
	to the -c option is specified, the current configuration byte
	is shown on the System.out.  If a parameter is specified, it
	is considered as a new configuration byte value and is written
	to the non-volatile configuration memory in DS1620.  The
	parameter may be specified either as a decimal number
	(consisting of decimal digits only) or as a hexadecimal number
	(prefixed by 0x.)  The value must be between 0 (0x00) and 255
	(0xFF) inclusive.  */

    public static void main ( final String[] argv )
    {
	try {
	    if ( argv.length == 0 ) {
		doVisual();
	    } else if ( argv.length == 1 && argv[0].startsWith( "-v" ) ) {
		doVisual();
	    } else if ( argv.length == 1 && argv[0].startsWith( "-t" ) ) {
		doTextural();
	    } else if ( argv.length == 1 && argv[0].startsWith( "-c" ) ) {
		doConfig();
	    } else if ( argv.length == 2 && argv[0].startsWith( "-c" ) ) {
		doConfig( argv[1] );
	    } else {
		doUsage();
	    }
	} catch ( Throwable e ) {
	    e.printStackTrace();
	} finally {
	    System.exit( 0 );
	}
    }

    /** Show the program usage.  */
    static void doUsage ()
    {
	System.out.println( "Thermo [-v | -t | -c [newconfig]]"   );
	System.out.println( "  -v  Run in visual mode (default)"  );
	System.out.println( "  -t  Run in text (console) mode"    );
	System.out.println( "  -c  Get/set config byte of DS1620" );
    }

    /** Show configuration byte (-c option.)  */
    static void doConfig ()
    {
	final Ds1620 ds = create1620();
	if ( ds != null ) {
	    try {
		System.out.println(
		    "DS1620 config: " + hex2( ds.getConfig() ) );
	    } finally {
		ds.close();
	    }
	}
    }

    /** Set a new configuration byte (-c param option.)
        <code>newConfStr</code> is the new config byte value, written
        either as a dicimal number or a hexadecimal number.  */
    static void doConfig ( final String newConfStr )
    {
	doConfig( newConfStr.startsWith( "0x" )
		  ? Integer.parseInt( newConfStr.substring( 2 ), 16 )
		  : Integer.parseInt( newConfStr ) );
    }

    /** Set a new configuration byte.  <code>newConfig</code> is the
        new config byte value.  */
    static void doConfig ( final int newConfig )
    {
	if ( newConfig < 0 || newConfig > 255 ) {
	    System.err.println(
		"Wrong config: must be an 8 bit number." );
	    return;
	}

	final Ds1620 ds = create1620();
	if ( ds != null ) {
	    try {
		final int oldConfig = ds.getConfig();
		ds.setConfig( newConfig );
		final int curConfig = ds.getConfig();
		System.out.println(
		    "DS1620 config set from: "
		    + hex2( oldConfig )
		    + " to: "
		    + hex2( curConfig ) );
	    } finally {
		ds.close();
	    }
	}
    }

    /** Run in text mode (-t.)  */
    static void doTextural ()
    {
	final Ds1620 ds = create1620();
	if ( ds != null ) {
	    try {
		ds.setOneShot( true );
		System.out.println(
		    ds.getTemperature() + " C"
		    + " ("
		    + ( Math.rint( ds.getHiResTemperature() * 100 ) / 100 )
		    + " C)" );
	    } finally {
		ds.close();
	    }
	}
	
    }

    /** The flag to control running of the program in visual mode.  It
        will keep running during <code>keepGoing</code> is true and
        will terminate when the variable is set to false.  */
    static volatile boolean keepGoing = true;

    /** Run in visual mode (-v.)  */
    static void doVisual ()
    {
	/* Setup the Bean to show the temperature.  */
	final ThermometerGraph tg = new ThermometerGraph();
	tg.setMaximum( 40 );	/* Maximum temperature is 40 C.  */
	tg.setMinimum(  0 );	/* Minimum temperature is  0 C.  */

	/* Setup the main (and only) window, containing a Bean.  */
	final Frame fr = new Frame();
	fr.addWindowListener( new WindowAdapter () {
		public void windowClosing ( final WindowEvent ev )
		{
		    ev.getWindow().dispose();
		}
		public void windowClosed ( final WindowEvent ev )
		{
		    keepGoing = false;
		}
	    });
	fr.setBackground( SystemColor.control );
	fr.add( tg );
	fr.pack();
	fr.show();

	/* Main measuring loop (with initialization.)  */
	final Ds1620 ds = create1620();
	if ( ds != null ) {
	    try {
		ds.setOneShot( true );
		while ( keepGoing ) {
		    tg.setValue( ds.getHiResTemperature() );
		    /* Stop execution for a while (5 seconds.)  One
                       temperature measurement takes about a second,
                       so the overall interval will be aprox. 6
                       secods.  We repeat one second sleep five times,
                       instead of single five seconds sleep, to
                       guarantee one second response to the program
                       termination from window close.
                       (ds.getHiResTemperature() takes about a seond,
                       so shorter sleep period is just inefficient.  */
		    for ( int w = 0; w < 5 && keepGoing; w++ ) {
			Thread.sleep( 1000 );
		    }
		}
	    } catch ( InterruptedException e ) {
		/* Return from this method.  */
	    } finally {
		ds.close();
	    }
	}
    }

    static Ds1620 create1620 ()
    {
	final UsbIo.Device usbio = UsbIo.getDevice();
	if ( usbio == null ) {
	    System.out.println( "no Morphy USBIO found." );
	    return null;
	}

	try {
	    /* Configure the USBIO as an SPI controller: Port#1/bit3
	       is clock, Port#0/bit7 is data, Port#0/bit5 is select
	       (active high).  Other pins are assumed to be
	       unconnected.  */
	    return new Ds1620(
		new UsbIoSpiDevice( usbio, 3, 7, 0x20, 0x00, 0x20 ) );
	} catch ( RuntimeException e ) {
	    usbio.close();
	    throw e;
	} catch ( Error e ) {
	    usbio.close();
	    throw e;
	}
    }

    static String hex2 ( final int val )
    {
	return "0x" + Integer.toHexString( val & 255 | 256 ).substring( 1 );
    }
}
