/*
USBIO Support Program
Copyright (C) 2001 by Masahiro Sekiguchi, all rights reserved.

Redistribution and use in source and/or binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote
   products derived from this software.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package xyzzy.awt;

import java.awt.*;
import java.awt.event.*;
import java.io.Serializable;

/** A visual Bean to show a Thermometer-like real-time bar graph.  */
public class ThermometerGraph
    extends Canvas
    implements Serializable
{
    /** Construct a ThermometerGraph with default properties.  */
    public ThermometerGraph ()
    {
	super();
	initialize();
    }

//  /** Construct a ThermometerGraph with the specified
//      GraphicsConfiguration.  */
//  public ThermometerGraph ( final GraphicsConfiguration gc )
//  {
//	super( gc );
//	initialize();
//  }

    /** Perform all necessary initialization.  This method is intended
        to be called from constructors only.  */
    protected void initialize ()
    {
	/* We need to know when the resize occured, since we need to
           recalculate gauges.  */
	addComponentListener( new ComponentAdapter () {
		public void componentResized ( final ComponentEvent ev )
		{
		    reset();
		}
	    } );
    }	

    /** The filledColor property.  */
    protected Color filledColor = Color.red;

    /** Set the filledColor.  FilledColor is the Color by which the
        bar (i.e., the filled area of the graph) is drawn.  */
    public void setFilledColor ( final Color color )
    {
	filledColor = color;
	repaint();
    }

    /** Get the current filledColor.  */
    public Color getFilledColor ()
    {
	return filledColor;
    }

    /** The hollowColor property.  */
    protected Color hollowColor = Color.white;

    /** Set the hollowColor Property.  hollowColor is the Color by
        which the area above the bar (i.e., the unfilled area of the
        graph) is drawn.  */
    public void setHollowColor ( final Color color )
    {
	hollowColor = color;
	repaint();
    }

    /** Get the current hollowColor.  */
    public Color getHollowColor ()
    {
	return hollowColor;
    }

    /** The value property.  */
    protected double value = 0;

    /** Set the value to be indicated to the user with this graph.  If
        the specified value is out of the current visible range, the
        written graph will be overflowed, but the internal value
        property holds the set value.  */
    public void setValue ( final double value )
    {
	this.value = value;
	final Graphics gr = getGraphics();
	if ( gr != null ) drawValue( gr );
    }
    
    /** Get the current value.  */
    public double getValue ()
    {
	return value;
    }

    /** The property maximum.  */
    protected double maximum = 100;

    /** Set the Maximum property.  Maximum is the value at which the
        bar reaches the upper most position.  Maximum must be larget
        than Minimum; otherwise the behaviour is not guaranteed.
        Value property may exceed Maximum, but the bar does not
        represent the Value in that case.  */
    public void setMaximum ( final double value )
    {
	maximum = value;
	if ( minimum >= value ) minimum = value - 0.5;
	reset();
    }

    /** Get the current value of Maximum property.  */
    public double getMaximum ()
    {
	return maximum;
    }

    /** The property minimum.  */
    protected double minimum = 0;

    /** Set the Minimum property.  Minimum is the value at which the
        bar reaches the lowermost position.  Minimum must be smaller
        than Maximum; otherwise the behaviour is not guaranteed.
        Value property may exceed Maximum, but the bar does not
        representthe Value in that case.  */
    public void setMinimum ( final double value )
    {
	minimum = value;
	if ( maximum <= value ) maximum = value + 0.5;
	reset();
    }

    /** Get the current value of Minimum property.  */
    public double getMinimum ()
    {
	return minimum;
    }

    /** The forcedToShowLabels property.  */
    protected boolean forcedToShowLabels = true;

    /** Set the forcedToShowLabels property.  If set to true, the bar
        is scaled so that both first and last labels are properly
        shown in the display area of this Bean.  If set to false, some
        label may be omitted and the bar is slightly extended to make
        more space.  Anyway, the bar has enough area to show both
        Minimum and Maximum values.  */
    public void setForcedToShowLabels ( final boolean forced )
    {
	forcedToShowLabels = forced;
    }

    /** Returns the forcedToShowLabels property.  */
    public boolean isForcedToShowLabels ()
    {
	return forcedToShowLabels;
    }

    /* The following three set* methods behave exactly as defined in
       Canvas.  No new semantic is added.  */

    public void setForeground ( final Color color )
    {
	super.setForeground( color );
	repaint();
    }

    public void setBackground ( final Color color )
    {
	super.setBackground( color );
	repaint();
    }

    public void setFont ( final Font font )
    {
	super.setFont( font );
	reset();
    }

    /** Width of the minimum visible size of this Bean.  */
    public static final int MIN_WIDTH = 61;

    /** Height of the minimum visible size of this Bean.  */
    public static final int MIN_HEIGHT = 81;

    /** Adjust the given <code>size</code> so that it is no smaller
        than the minimum size of this Bean.  The adjusted Dimension is
        returned.  */
    protected static Dimension ensureMinimum ( final Dimension size )
    {
	if ( size.width  < MIN_WIDTH  ) size.width  = MIN_WIDTH;
	if ( size.height < MIN_HEIGHT ) size.height = MIN_HEIGHT;
	return size;
    }

    /** Returns the minimum size of this Bean as a visible component.  */
    public Dimension getMinimumSize ()
    {
	return new Dimension( MIN_WIDTH, MIN_HEIGHT );
    }

    /** Returns the preferred size of this Bean as a visible
        component.  */
    public Dimension getPreferredSize ()
    {
	return ensureMinimum( super.getPreferredSize() );
    }

    /** Indicate that the layout metrics has been changed since last
        redraw.  This variable is used to optimize the operation.  */
    protected volatile boolean metricsChanged = true;

    /** Force repainting this Bean.  */
    protected void reset ()
    {
	metricsChanged = true;
	repaint();
    }

    /* Various internal metrics parameters.  Mostly undocumented.
       Sorry.  */

    protected int barL, barR;
    protected int barT, barB;
    protected int potX, potY, potR;
    protected int minPos, maxPos, incPos;
    protected int minLab, maxLab, incLab, extLab;
    protected double valB, valT;
    protected String[] labList = new String[0];
    protected int[] offList;

    /** Find the set of new metrics to be used to draw this Bean.  The
        Graphics <code>gr</code> given as an argument is used to
        extract several font metrics values.  Otherwise the operation
        of findMetrics is independent from any Graphics.  */
    protected void findMetrics ( final Graphics gr )
    {
	/* We just keep the old metrics unless something has been
           changed.  */
	if ( !metricsChanged ) return;
	metricsChanged = false;

	/* Prepare for the font metrics to calculate the size of labels.  */
	final FontMetrics fm = gr.getFontMetrics( getFont() );

	/* Calculate base metrics.  */
	final Dimension dim = ensureMinimum( getSize() );
	final int w = dim.width;
	final int h = dim.height;

	/* Determine the size (radius) of the pot.  */
	int pr = w / 3;
	if ( pr > h / 8 ) pr = h / 8;

	/* Estimate the hight of the label.  This is just a rough
           estimate.  FIXME.  */
	int u = fm.getAscent() + 1;
	
	/* Determine the size of the bar.  */
	int bh = h - u - ( 2 * pr - 1 );
	if ( bh < 3 ) bh = 3;
	int bw = pr;
	if ( ( bw & 1 ) == 0 ) bw -= 1;
	if ( bw < 3 ) bw = 3;

	/* Find the value-drawing metrics.  */
	barL = ( w - bw ) / 2;
	barR = barL + bw - 1;
	barT = u;
	barB = h - u - ( 2 * pr );
	potX = barL + ( bw / 2 );
	potY = barB + pr;
	potR = pr;

	/* Estimate the expected number of gauges on the bar.  */
	final int fh = fm.getHeight();
	int g = ( barB - barT + 1 ) / fh;
	if ( g <  1 ) g = 1;
	if ( g > 20 ) g = 20;

	/* Estimate the gauge unit.  */
	double gu = ( maximum - minimum ) / g;
	int e = 0;
	double er = 1;
	while ( gu > 10 ) { e++; er *= 10; gu /= 10; }
	while ( gu <= 1 ) { e--; er *= 10; gu *= 10; }

	/* We (human beings) use decimal number system, so the gauge
           unit should be a divisor of ten (after scaled with some
           powers of 10.)  */
	int gi;
	if ( gu <= 2 ) {
	    gi = 2;
	} else if ( gu <= 5 ) {
	    gi = 5;
	} else {
	    gi = 10;
	}
	gu = ( e >= 0 ) ? ( gi * er ) : ( gi / er );

	/* Adjust (lower) the minimum value on an integral multiple of
           the gauge unit.  */
	if ( e >= 0 ) {
	    valB = Math.floor( minimum / er / gi ) * gi * er;
	} else {
	    valB = Math.floor( minimum * er / gi ) * gi / er;
	}

	/* Re-evaluate the expected number of gauges based on the
           calculated gauge unit and the adjusted minimum value.  */
	g = ( int )Math.ceil( ( maximum - valB ) / gu );

	/* Now comes the most complecated part.  We want gauges be
           equally distributed on the display.  I.e., the distance of
           gauges, measured in pixels on the display, must be exactly
           same.  So we have to round the distance.  However, it is
           not straightforward to determine we should round up or
           down.  To find the optimal distance, we try loose
           allocation first, and if it doesn't fit in the given
           display area, we use tight alloation.  When seeing the
           fitness, we count the gap above the (provisional) top of
           the bar that was reserved for the topmost label, if
           appropriate.  */
	incPos = ( barB - barT + g - 1 ) / g;
	int bn = ( int )Math.ceil( ( maximum - valB ) / gu * incPos );
	if ( barB - bn >= barT) {
	    /* OK.  No problem.  */
	} else if ( !isForcedToShowLabels() && barB - bn > 0 ) {
	    /* It fits in the display area.  Extend the bar drawing
               area to cover needed range.  */
	    barT = barB - bn;
	} else if ( incPos > 1 ) {
	    /* We need to use a tight gauge in this case.  */
	    incPos -= 1;
	} else {
	    /* We should not come here, since we have sufficiently
               large MIN_HEIGHT value... */
	    throw new RuntimeException( "Oops" );
	}

	/* Calculate other gauge metrics.  */
	minPos = barB; /* This is fixed by design.  */
	maxPos = barB - ( barB - barT ) / incPos * incPos;
	valT = ( double )( barB - barT ) / incPos * gu + valB;

	/* Generate guage label texts.  */
	g = ( minPos - maxPos ) / incPos + 1;
	if ( labList.length != g ) {
	    labList = new String[g];
	    offList = new int[g];
	}
	int maxWidth = 0;
	for ( int n = 0; n < labList.length; n++ ) {
	    labList[n] = String.valueOf( valB + n * gu );
	    offList[n] = fm.stringWidth( labList[n] );
	    if ( offList[n] > maxWidth ) maxWidth = offList[n];
	}
	for ( int n = 0; n < offList.length; n++ ) {
	    offList[n] = maxWidth - offList[n];
	}
    }

    /** Draw the current Value on the bar using given Graphics.  */
    protected void drawValue ( final Graphics gr )
    {
//	if ( metricsChanged ) return;

	/* Make the plotting position in the visible bar area, if the
           value exceeds the range in Minimum and Maximum.  */
	final int plot;
	if ( value <= valB ) {
	    plot = barB;
	} else if ( value >= valT ) {
	    plot = barT;
	} else {
	    plot = barB - ( int )( ( value - valB )
				   / ( valT - valB ) * ( barB - barT ) );
	}

	/* Draw it.  */
	gr.setPaintMode();
	gr.setColor( getHollowColor() );
	gr.fillRect( barL + 1, barT, barR - barL - 1, plot - barT );
	gr.setColor( getFilledColor() );
	gr.fillRect( barL + 1, plot, barR - barL - 1, potY - plot + 1 );
    }

    /** Draw the (outline and other fixed parts of) bar.  This method
        destroys the variable part, so drawValue should immediately be
        called.  */
    protected void drawBar ( final Graphics gr )
    {
	/* Note that, surprisingly, fillOval and drawOval draws
           _different_ sizes of ovals if exactly the same parameters
           are give.  :-( */
	gr.setPaintMode();
	gr.setColor( getForeground() );
	gr.drawLine( barL, barT, barL, potY );
	gr.drawLine( barR, barT, barR, potY );
	gr.setColor( getFilledColor() );
	gr.fillOval( potX - potR, potY - potR, potR * 2 + 1, potR * 2 + 1 );
	gr.setColor( getForeground() );
	gr.drawOval( potX - potR, potY - potR, potR * 2, potR * 2 );
	for ( int n = 0, p = minPos; n < labList.length; n++, p -= incPos ) {
	    gr.drawLine( barR + 1, p, barR + 10 /* XXX */, p );
	    gr.drawString( labList[n], barR + 10 /* XXX */ + offList[n], p );
	}
    }

    public void paint ( final Graphics gr )
    {
	super.paint( gr );
	findMetrics( gr );
	drawBar( gr );
	drawValue( gr );
    }
}
