/*
USBIO Support Program
Copyright (C) 2001 by Masahiro Sekiguchi, all rights reserved.

Redistribution and use in source and/or binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote
   products derived from this software.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package xyzzy.hardware;

/** The abstract base class for all device classes based on the DMF.  */
abstract public class DeviceObject
{
    /** The object to synchronize against when multiple operations on
        this device must be performed atomic.  If a device provides
        some interface functionality over another device, the lock
        object of the interface object is usually copied from the
        underlying device object.  Note that this strategy of having
        separate lock object has pros and cons; When underlying device
        provides two or more totally independent _channels_, locking
        against underlying device unnecessary blocking of other client
        devices.  We may eventually reconsider this point.  FIXME.  */
    protected final Object lock;

    /** Constract a new DeviceObject using itself as a lock object.
        This constractor is intended to be used by self-standing
        devices.  */
    protected DeviceObject ()
    {
	this.lock = this;
    }

    /** Constract a new DeviceObject using specified lock object.
        This constrator is intended to be used by devices that work
        on top of another devices.  */
    protected DeviceObject ( final Object lock )
    {
	this.lock = lock;
    }

    /** Get an object to synchronize against when two or more
        operations (methods) are to be performed atomic.  Suppose
        <code>dev</code> is an instance of (a subclass of)
        DeviceObject.  If your application needs an atomic operations,
        you should write <code>synchronized (dev.getLockObject()) {
        ... }</code> instead of <code>synchronized (dev) {
        ... }</code>.  */
    public final Object getLockObject ()
    {
	return lock;
    }

    /** Make this device quitted from operation.  Any valuable
        resources allocated for this object/device will be released.
        After calling close(), a DeviceObject object is unfunctional.
        The close() method may be called two or more times for an
        object, although second or later call does nothing useful.  */
    abstract public void close ();
}

