/*
USBIO Support Program
Copyright (C) 2001 by Masahiro Sekiguchi, all rights reserved.

Redistribution and use in source and/or binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote
   products derived from this software.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package xyzzy.hardware;

/** Supports Dallas Semicondactor DS1620 digital thermometer.

    DS1620 can be programmed to one of two modes: continuous mode and
    one-shot mode.  The Ds1620 class also provides two operations:
    continuous operation and one-shot operation.  The hardware mode
    and the software operation are similar (and in fact related) but
    not same.  Ds1620 support three combinations of modes and
    operations out of possible three.  The following table summarizes
    the cases:

    <table>
    <tr><th>Hardware Mode</th><th>Software Operation</th><th>Allowed?</th></tr>
    <tr><td>continuous</td>   <td>continuous</td>        <td>yes</td>     </tr>
    <tr><td>continuous</td>   <td>one-shot</td>          <td>yes</td>     </tr>
    <tr><td>one-shot</td>     <td>continuous</td>        <td>no</td>      </tr>
    <tr><td>one-shot</td>     <td>one-shot</td>          <td>yes</td>     </tr>
    </table>

    The hardware mode can be set by writing into the chip's
    configuration register.  The Ds1620 class provides that function
    through setConfig(int) method.  The software operation can be set
    by setOneShot(boolean) method.  When invokes, setOneShot(boolean)
    verifies the underlying hardware's mode, and, if the hardware mode
    and the specified software operation causes an disallowed
    combination, an Exception is thrown and the operation will be
    adjusted to cause an allowed combination.  However, setconfig(int)
    doesn't care about the software operation, and it is possible to
    make an disallowed combination as a result.  It is the
    application's responsibility not to do so.<br>

    A Ds1620 object can be set to continuous operation only if the
    hardware is in continuous mode.  If a Ds1620 object is
    (successfully) set to continuous operation, it utilizes the
    hardware's continuous mode to let it measure the temperature
    continuously.  As a result, the getTemperature() method returns
    very quickly with the latest temperature.  As a drawback,
    getHiResTemperature() doesn't work on continuous operation.  (It
    is the hardware limitation.)<br>

    You can put a Ds1620 object into one-shot operation regardless of
    the hardware mode.  In one-shot operation, even if the hardware is
    in continuous mode, the Temperature measurement unit will usually
    be stopped and will be activated only once per a call to
    getTemperature() (or getHiResTemperature()) method.  As good news,
    getHiResTemperature() works with one-shot operation, and the power
    consumption by the DS1620 chip will be minimized.  Bad news is
    that getTemperature() takes several milliseconds to complete in
    one-shot operation, since The hardware's Temperature measurement
    activity will be initiated after the call.

    */
public class Ds1620
    extends SpiTarget
{
    /* DS1620 command codes, as in the datasheet (072099).  */
    protected final int READ_TEMPERATURE = 0xAA;
    protected final int READ_COUNTER = 0xA0;
    protected final int READ_SLOPE = 0xA9;
    protected final int START_CONVERT_T = 0xEE;
    protected final int STOP_CONVERT_T = 0x22;
    protected final int WRITE_TH = 0x01;
    protected final int WRITE_TL = 0x02;
    protected final int READ_TH = 0xA1;
    protected final int READ_TL = 0xA2;
    protected final int WRITE_CONFIG = 0x0C;
    protected final int READ_CONFIG = 0xAC;

    /* Definition for config/status register bits as in DS1620
       datasheet.  */

    /** Config/status register bit mask for conversion (measurement)
        done.  */
    public final int DONE = 0x80;

    /** Config/status register bit mask for temperature high flag.  */
    public final int THF = 0x40;

    /** Config/status register bit mask for temperature low flag.  */
    public final int THL = 0x20;

    /** Config/status register bit mask for non-volatile memory busy.  */
    public final int NVB = 0x10;

    /** Config/status register bit mask for CPU mode (as opposed to
        stand-alone mode.)  */
    public final int CPU = 0x02;

    /** Config/status register bit mask for one-shot mode (as opposed
        to continuous mode.)  */
    public final int OS = 0x01;

    /** Indicates that this object is in one-shot operation.  The
        default is true, meaning one-shot operation.  Note that
        various parts of the class implemention depends on this
        default, and changing the default to true will need a lot of
        rewriting to the Ds1620 source code.  */
    protected boolean oneshot = false;

    /** Create a DS1620 object that is assumed to be connected through
        an SPI contoller device represented by <code>spi</code>.  */
    public Ds1620 ( final SpiDevice spi )
    {
	super( spi, true );
    }    

    /** Put this object in the continuous operation if the
        <code>oneshot</code> is false, put in the one-shot operation
        otherwise.  The hardware must be set to continuous mode before
        calling setOneShot(false) to set to continuous operation.  If
        this method is called with its argument false and the hardware
        is in one-shot mode, the object is put in one-shot operation
        and an Exception is thrown.  */
    public void setOneShot ( final boolean oneshot )
    {
	if ( !oneshot && ( getConfig() & OS ) != 0 ) {
	    /* continuous operation is requested when The hardware is
               in one-shot mode.  This combination is unsupportable
               and disallowed.  */
	    this.oneshot = true;
	    throw new IllegalArgumentException(
		"continuous operation requested when in one-shot mode" );
	}
	/* Act as requested.  */
	this.oneshot = oneshot;
	spi.exec( oneshot ? STOP_CONVERT_T : START_CONVERT_T );
    }

    /** Returns whether this object is in the oneshot operation mode.  */
    public boolean isOneShot ()
    {
	return oneshot;
    }

    /** Read the chip's configuration/status register and return the
        value.  */
    public int getConfig ()
    {
	return ( int )spi.execRead( READ_CONFIG, 8 );
    }
	
    /** Change the chip's configuration/status register.  Note that,
        although the bit pattern specified with <code>config</code>
        will be passed to the chip as-is, some of them may be ignored.
        See hardware datasheet for details.  Changing the non-volatile
        config bit takes several milliseconds.  The method doesn't
        return until this underlying write cycle to complete.  */
    public void setConfig ( final int config )
    {
	synchronized ( getLockObject() ) {

	    /* Write the specified config byte.  */
	    spi.execWrite( WRITE_CONFIG, config, 8 );

	    /* Wait until the chip completes its underlying write
	       cycle to the internal EEPROM, i.e., until the NVB bit
	       in the status register turns off.  */
	    for (;;) {
		sleep( 1 );
		if ( ( getConfig() & NVB ) == 0 ) break;
	    }
	}
    }

    /** Get a measured temperature value as its raw form.  The raw
        form is a 9 bit signed integer (as promoted to 32 bit integer
        for Java compatibility) representing a temperature in 0.5
        Celsius degree.  */
    public int getRawTemperature ()
    {
	synchronized ( getLockObject() ) {

	    /* If we are in one-shot operation, start measuring
	       temperature, then stop it immediately, in case the chip
	       is in continuous mode.  Then wait for th measurement to
	       complete.  */
	    if ( oneshot ) {

		spi.exec( START_CONVERT_T );
		sleep( 1 );
		spi.exec( STOP_CONVERT_T );

		/* Wait for the measurement to complete.  */
		for (;;) {
		    sleep( 100 );
		    /* Done? */
		    if ( ( getConfig() & DONE ) != 0 ) break;
		}
	    }

	    /* In continuous mode, we need 1 second delay after either
               power-up or re-configuration.  Unfortunately, we
               currently have no way to know this required 1 second
               have been elapsed.  In this version, we just read the
               temperature.  FIXME.  */

	    /* Issue Read Temperature command and return the result
               after sign extend to 32 bit.  */
	    final int val = ( int )spi.execRead( READ_TEMPERATURE, 9 );
	    return ( val ^ 0x100 ) - 0x100;	
	}
    }

    /** Get a measured temperature value.  The returned double number
        is a temperature in Celsius degree.  It has 0.5C resolution
        and 0.5C accuracy.  See the hardware documentation for
        details.  When in one-shot operation, the method issues START
        CONVERT T command to initiate the measurement cycle, wait for
        the cycle to complete, and read the result out of the chip.
        When in continuous operation, the method just read the latest
        result from the chip.  */
    public double getTemperature ()
    {
	return ( double )getRawTemperature() / 2;
    }

    /** Get a high resolution temperature value.  This method uses an
        interpolation technique described in the hardware document to
        get the temperature in extended resolution, which is expected
        to have less than 0.1C.  (It doesn't mean the measured value
        has more accuracy than getTemperature().  See hardware
        document for details.)  getHiResTemperature() works only when
        in one-shot operation.  If this method is called when it is
        continuous operation, an exception is thrown.  */
    public double getHiResTemperature ()
    {
	synchronized ( getLockObject() ) {
	    if ( !oneshot ) {
		throw new RuntimeException(
		    "HiRes Temperature unavailable on continuous operation" );
	    }

	    /* Read the base temperature value.  The LSB (as in its
               raw form) must be removed before interporation.  */
	    final double t = getRawTemperature() >> 1;

	    /* Read the underlying counters.  */
	    final double r = ( int )spi.execRead( READ_COUNTER, 9 );
	    final double k = ( int )spi.execRead( READ_SLOPE, 9 );
	    
	    /* Interporate the base temperature with counters.  */
	    return t + 0.75 - r / k;
	}
    }

    /** Just stop execution for the specified milliseconds.  This is
        just like the java.lang.Thread.sleep(int), except that
        Ds1620.sleep(int) never throws InterruptedException.  */
    protected void sleep ( int millis )
    {
	try {
	    Thread.sleep( millis );
	} catch ( InterruptedException e ) {
	}
    }
}
