/*
USBIO Support Program
Copyright (C) 2001 by Masahiro Sekiguchi, all rights reserved.

Redistribution and use in source and/or binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote
   products derived from this software.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package xyzzy.hardware;

/** abstract class to provide common codebase for Microwire (TM)
    compatible Serial EEPROMs.  (aka 93 series.)  Use one of
    subclasses to access to an actual memory.  Some 93 series device
    support byte oriented organization as well as standard word
    oriented organization.  The target chip, however, must be wired as
    word-oriented organization, when to be controlled by one of
    subclasses of MWMemory class.  This class doesn't support those
    chips wired to byte-oriented organization.  */
abstract public class MWMemory
    extends SpiTarget
{
    final protected int addrLen;

    public MWMemory ( final SpiDevice spi, final int addrLen )
    {
	/* Microwire SEEPROM uses MSB First format.  */
	super( spi, false );

	/* Prepare for various constants.  */
	this.addrLen = addrLen;
    }

    /** Change the 93Cx6's write enable mode.  When called with
        <code>enable</code> true, the chip will be write-enabled.
        When called with <code>enable</code> false, the chip will be
        write-protected.  The MWMemory class assumes no default,
        although the datasheet says the chip is in write-protected
        mode upon power-on.  The MWMemory class neither provides
        <i>query</i> method for the write enable mode, since the chip
        has no such function.  */
    public void setWriteEnabled ( final boolean enable )
    {
	spi.execute( ( enable ? 19 : 16 ) << ( addrLen - 2 ), addrLen + 3, 0 );
    }

    /** Read a word (16 bit data) at the specified address.  The
        address must be a 6 bit number.  */
    public int read ( final int addr )
    {
	if ( ( addr >> addrLen ) != 0 ) {
	    throw new IndexOutOfBoundsException( "Address: " + addr );
	}
	/* Send the output bits and receive input bits.  Note that the
           data bits are preceeded by an isolation bit, so we actually
           need to read 17 bits, although it is guaranteed to be a
           zero and all transmissions between 93 series chips are done
           as MSB first, so we don't need to strip it from the return
           value.  */
	return ( int )spi.execute( ( 6 << addrLen ) | addr,
				   addrLen + 3, 17 );
    }

    /** Write a word (16 bit data) into the specified address.  The
        address must be a 6 bit number.  The call doesn't return until
        underlying write cycle on the chip completes.  */
    public void write ( final int addr, final int value )
	throws InterruptedException
    {
	if ( ( addr >> addrLen ) != 0 ) {
	    throw new IndexOutOfBoundsException( "Address: " + addr );
	}
	if ( ( value & ~65535 ) != 0 ) {
	    throw new IllegalArgumentException(
		"Value exceeds 16 bits: " + value );
	}
	synchronized ( getLockObject() ) {
	    /* Send the output bits that consist of: START, COMMAND,
	       ADDRESS, and DATA.  */
	    spi.execute( ( 5 << ( addrLen + 16 ) ) | ( addr << 16 ) | value,
			 addrLen + 19, 0 );
	    /* Wait for the write cycle to complete.  */
	    waitForReady();
	}
    }

    /** Erase a word (16 bit data) at the specified address.  The
        address must be a 6 bit number.  After erase, the memory cell
        at the address becomes all 1s.  This operation is logically
        equivallent to writing 0xFFFF, but expected to be faster.  The
        call doesn't return until underlying write cycle on the chip
        completes.  */
    public void erase ( final int addr )
	throws InterruptedException
    {
	if ( ( addr >> addrLen ) != 0 ) {
	    throw new IndexOutOfBoundsException( "Address: " + addr );
	}
	synchronized ( getLockObject() ) {
	    /* Send the output bits that consist of: START(1),
	       COMMAND(2), ADDRESS(6), DATA(16).  Note that the
	       preceeding isolation bit is silently ignored since it
	       overlaps with the last output bit.  Also note that all
	       transmissions are done as MSB first.  */
	    spi.execute( ( 7 << addrLen ) | addr, addrLen + 3, 0 );
	    /* Wait for the write cycle to complete.  */
	    waitForReady();
	}
    }

    /** Erase the entire chip.  The call doesn't return until
        underlying write cycle on the chip completes.  */
    public void eraseAll ()
	throws InterruptedException
    {
	synchronized ( getLockObject() ) {
	    /* Send the output bits that consist of: START(1),
	       COMMAND(2), ADDRESS(6), DATA(16).  Note that the
	       preceeding isolation bit is silently ignored since it
	       overlaps with the last output bit.  Also note that all
	       transmissions are done as MSB first.  */
	    spi.execute( 18 << ( addrLen - 2 ), addrLen + 3, 0 );
	    /* Wait for the write cycle to complete.  */
	    waitForReady();
	}
    }

    /** Wait until the target 93Cx6 is READY.  This method is intended
        to be called after write or erase operation.  The caller must
        make sure to have a monitor of the lock object before calling
        this method.  */
    protected void waitForReady ()
	throws InterruptedException
    {
	/* Just poll the DO line of the chip and wait until it comes
	   high, i.e., READY status.  */
	while ( spi.poll() == 0 ) {
	    /* Wait another 1ms.  */
	    Thread.sleep( 1 );
	}
    }
}
