/*
USBIO Support Program
Copyright (C) 2001 by Masahiro Sekiguchi, all rights reserved.

Redistribution and use in source and/or binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote
   products derived from this software.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package xyzzy.hardware;

import java.io.InputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;

public final class MachDepManager
{
    private static final String PROPERTY_FILE = "MachDep.properties";
    private static final String PROPERTY_PREFIX = "MachDep.package.";
    private static final int PROPERTY_PREFIX_LENGTH = PROPERTY_PREFIX.length();
    private static final int MAX_MACHDEP_PRIORITY = 10;

    private static final String[] packages;
    static {
	final Properties prop;
	try {
	    prop = Utils.getResourceAsProperties( MachDepManager.class,
						  PROPERTY_FILE );
	} catch ( IOException e ) {
	    throw new RuntimeException( e.toString() );
	}
	    
	final Vector vec = new Vector();
	final Enumeration names = prop.propertyNames();
	while ( names.hasMoreElements() ) {
	    final String name = ( String )names.nextElement();
	    if ( name.startsWith( PROPERTY_PREFIX ) ) {
		final String priority
		    = name.substring( PROPERTY_PREFIX_LENGTH );
		final int p;
		try {
		    p = Integer.parseInt( priority );
		} catch ( NumberFormatException e ) {
		    throw new RuntimeException(
			e.toString() + " (" + name + ")" ); 
		}
		if ( p > MAX_MACHDEP_PRIORITY ) {
		    throw new RuntimeException(
			"too large priority: " + name );
		}
		if ( vec.size() < p + 1 ) vec.setSize( p + 1 );
		vec.setElementAt( prop.getProperty( name ), p );
	    }
	}

	int count = 0;
	for ( int i = 0; i < vec.size(); i++ ) {
	    String s = ( String )vec.elementAt( i );
	    if ( s != null ) {
		if ( !s.endsWith( "." ) ) s += ".";
		vec.setElementAt( s, count++ );
	    }
	}
	vec.setSize( count );
	packages = new String[count];
	vec.copyInto( packages );
    }

    public static Object getMachDep ( final Class definitive )
    {
	/* Find the unqualified (base) name of the definitive class.
           Each implementing class is assumed to have the same
           unqualified name.  */
	final String fqn = definitive.getName();
	final int dot = fqn.lastIndexOf( '.' );
	final String name = ( dot < 0 ? fqn : fqn.substring( dot + 1 ) );

	/* Look up the implementing class for the definitive class.  */
	for ( int i = 0; i < packages.length; i++ ) {
	    try {
		final Class cl = Class.forName( packages[i] + name );
		if ( !MachDep.class.isAssignableFrom( cl ) ) continue;
		if ( !definitive.isAssignableFrom( cl ) ) continue;
		return cl.newInstance();
	    } catch ( Throwable e ) {
		continue;
	    }
	}

	/* No machine dependent class could be instanciated.  */
	throw new NoClassDefFoundError( "MachDep: " + name );
    }
}


