/*
USBIO Support Program
Copyright (C) 2001 by Masahiro Sekiguchi, all rights reserved.

Redistribution and use in source and/or binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote
   products derived from this software.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package xyzzy.hardware;

/** Abstract base class for all SPI and Microwire controller devices,
    which are sometimes referred to as _bus_.  */
abstract public class SpiDevice
    extends DeviceObject
{
    /** Specifies the bit serialization order of integral values on
       the SPI data line.  */
    protected boolean lsbFirst = true;

    /** Create a self-standing SPI device object.  The bit
        serialization order is assumed to be LSB-first.  */
    protected SpiDevice ()
    {
	super();
    }

    /** Create an SPI device object that runs on another device object
        whose lock object is <code>lock</code>.  The bit serialization
        order is assumed to be LSB-first.  */
    protected SpiDevice ( final Object lock )
    {
	super( lock );
    }

    /** Create an SPI device object that runs on another device object
        whose lock object is <code>lock</code>.  The bit serialization
        order is given by <code>lsbFirst</code>.  */
    protected SpiDevice ( final Object lock, final boolean lsbFirst )
    {
	super( lock );
	setLsbFirst( lsbFirst );
    }

    public void setLsbFirst ( final boolean lsbFirst )
    {
	this.lsbFirst = lsbFirst;
    }

    public boolean isLsbFirst ()
    {
	return lsbFirst;
    }

    /** Issue a command byte only to SPI device.  */
    public final void exec ( final int cmd )
    {
	execute( cmd, 8, 0 );
    }

    /** Issue a command byte and read several bits from SPI device.  */
    public final long execRead ( final int cmd, final int len )
    {
	return execute( cmd, 8, len );
    }

    /** Issue a command byte and write argument bits to SPI device.  */
    public final void execWrite ( final int cmd,
				  final long arg, final int len )
    {
	if ( isLsbFirst() ) {
	    execute( cmd | ( arg << 8 ), len + 8, 0 );
	} else {
	    execute( ( cmd << len ) | arg, len + 8, 0 );
	}
    }

    /** Issue a command byte, write argument bits, then read other
        bits from SPI device.  */
    public final long execWriteRead ( final int cmd,
				      final long arg, final int argLen,
				      final int readLen )
    {
	if ( isLsbFirst() ) {
	    return execute( cmd | ( arg << 8 ), argLen + 8, readLen );
	} else {
	    return execute( ( cmd << argLen ) | arg, argLen + 8, readLen );
	}
    }

    /** Issue a command to an SPI chip, i.e., activate select, clock
        in (send) several data bits (including command code), and
        clock out (receive) several data bits.  */
    abstract public long execute ( long outbits, int outlen, int inlen );

    /** Just sence the DO line of the SPI chip, i.e., activate select,
        <em>without</em> suppling clock.  */
    abstract public int poll ();

    protected void exception ( final String msg )
    {
	throw new SpiException( msg );
    }

    public static class SpiException
        extends RuntimeException
    {
	public SpiException ()
	{
	    super();
	}

	public SpiException ( final String msg )
	{
	    super( msg );
	}
    }
}
