/*
USBIO Support Program
Copyright (C) 2001 by Masahiro Sekiguchi, all rights reserved.

Redistribution and use in source and/or binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote
   products derived from this software.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package xyzzy.hardware;

import java.util.Enumeration;
import java.util.NoSuchElementException;

/** UsbIo is the topmost user-visible class to manipulate Morphy USBIO
    devices.<br>

    This is an abstract class, so no application program can create
    any instances of it directly.  Moreover, under no circumstances,
    should any instances of any subclasses of this class be directly
    instanciated by application programs.  Such an instance is create
    by UsbIo internally (in its static initializer.)<br>

    This class facilitates two different ways.<br>

    For application programs, this class provides several public
    static methods to be called by applications.  They give ways to
    access UsbIo devices.<br>

    For class library implementation, this class acts as the base
    class for all platform-dependent classes to implement low-level
    USB support functions.  One of the subclasses will be selected and
    instanciated at run time to act something like a pier to this
    class.  */

abstract public class UsbIo
{
    /* List of constants that depend on USB IO device hardware
       specification.  */

    /** Command byte to write data to Port #0.  */
    public static final int WRITE0 = 0x01;

    /** Command byte to write data to Port #1.  */
    public static final int WRITE1 = 0x02;

    /** Command byte to read data from Port #0.  */
    public static final int READ0 = 0x03;

    /** Command byte to read data from Port #1.  */
    public static final int READ1 = 0x04;

    /** Command byte to write data to Port #0 with a strobe on P1.0.  */
    public static final int WRITE0_P0 = 0x10;

    /** Command byte to write data to Port #0 with a strobe on P1.1.  */
    public static final int WRITE0_P1 = 0x11;

    /** Command byte to write data to Port #0 with a strobe on P1.2.  */
    public static final int WRITE0_P2 = 0x12;

    /** Command byte to write data to Port #0 with a strobe on P1.3.  */
    public static final int WRITE0_P3 = 0x13;

    /** Command byte to read data from Port #0 during a strobe on P1.0.  */
    public static final int READ0_P0 = 0x14;

    /** Command byte to read data from Port #0 during a strobe on P1.1.  */
    public static final int READ0_P1 = 0x15;

    /** Command byte to read data from Port #0 during a strobe on P1.2.  */
    public static final int READ0_P2 = 0x16;

    /** Command byte to read data from Port #0 during a strobe on P1.3.  */
    public static final int READ0_P3 = 0x17;

    /** Pseudo command modifier to request asynchronous operation,
        i.e., send the request only (without waiting for a result.)
        This modifier may be ignored on some platform.  */
    public static final int ASYNC = 0x1000;

    /** Pseudo command code to receive (through the Interrupt pipe)
        only.  */
    public static final int RECEIVE = 0x100;

    /** The pier instance of platform-dependent UsbIo implementation
        class.  An appropriate object is instanciated upon class
        loading of UsbIo and kept forever.  */
    protected static final UsbIo theInstance
        = ( UsbIo )MachDepManager.getMachDep( UsbIo.class );

    /** A kind of a debug flag.  Read from a dedicated properties file.  */
    private static final boolean NO_DOUBLE_CHECK
        = Boolean.getBoolean( "xyzzy.hardware.UsbIo.NO_DOUBLE_CHECK" );

    /** Another debug flag.  */
    private static final boolean VERBOSE_DOUBLE_CHECK
        = Boolean.getBoolean( "xyzzy.hardware.UsbIo.VERBOSE_DOUBLE_CHECK" );

    /** Returns a USBIO Device object.  If two or more USBIO devices
        are attached to the system, one of them is arbitrarily chosen.
        If there are no USBIO device, the method returns a null.
        getDevice() is intended to be used by systems that allow only
        one USBIO device at a time.  */
    public static Device getDevice ()
    {
	final Object[] deviceHandles = theInstance.findDevices();
	if ( deviceHandles.length > 0 ) {
	    return theInstance.createDevice( deviceHandles[0] );
	} else {
	    return null;
	}
    }

    /** Returns an enumeration of all USBIO Device objects on the
        system.  The returned Enumeration object is an instance of
        UsbIo.DeviceEnumeration.  */
    public static Enumeration allDevices ()
    {
	return new DeviceEnumeration();
    }

    /** The Enumeration that is returned by devices().  */
    protected static class DeviceEnumeration
        implements Enumeration
    {
	/** The list of device handles that represent devices to be
            enumerated.  */
	protected final Object[] deviceHandles;

	/** The index to the next element to be enumerated.  */
	protected int next;

	/** Create an enumeration of devices from a list of device
            handles.  */
	protected DeviceEnumeration ()
	{
	    deviceHandles = theInstance.findDevices();
	    next = 0;
	}

	/** True if there are more Device's, as always.  */
	public boolean hasMoreElements ()
	{
	    return next < deviceHandles.length;
	}

	/** Returns a USBIO device attached on the system.  Although
            the returned object is defined as Object, it is guaranteed
            to be an instance of UsbIo.Device.  The returned Device
            object is ready to use, i.e., all necessary preparation
            jobs are done.  That means the caller is responsible to
            call close() against the object before releasing it.  (I
            agree that this design is brain damaged, but this is the
            course of Java...  FIXME.  */
	public Object nextElement ()
	    throws NoSuchElementException
	{
	    if ( next < deviceHandles.length ) {
		return theInstance.createDevice( deviceHandles[next++] );
	    } else {
		throw new NoSuchElementException();
	    }
	}
    }

    /** Create a Device object from a device handle.  As a default
        behaviour, UsbIo.createDevice instanciate the Device class and
        returns it.  Subclasses may override it.  */
    protected Device createDevice ( final Object handle )
    {
	return new Device( handle );
    }

    /** Class for USBIO Device.  The UsbIo.Device provides a
        reasonable set of default implementation.  Any implementation
        is allowed to subclass Device for its own purpose.
        Application programs are requested to use UsbIo.Device as
        placeholder, always.  */
    public static class Device
        extends DeviceObject
    {
	protected int fd;

	protected Device ( final Object handle )
	{
	    fd = theInstance.openDevice( handle );
	    if ( fd < 0 ) {
		throw new UsbIoException( "openDevice failed: " + handle );
	    }
	}

	/** Releases any valuable resources associated with this
            object.  After calling close(), this object is
            unfunctional.  It is the application's responsibility to
            ensure that this method is called before garbage collector
            reclaims this object.  Failure to do so will cause some
            system resource to leak forever.  (As you know, this is
            the famous stupidity of Java.)  */
	public void close ()
	{
	    if ( fd >= 0 ) {
		theInstance.closeDevice( fd );
		fd = -1;
	    }
	}

	/* Some variables that are used only in invoke below but are
           declared outside of it due to various circumstances.  (If
           this were C or C++, they are declared as invoke's local
           variables.  */

	/** A request serial number.  It is used to match _read_
            commands with their result.  It is incremented for each
            message roundtrip, to distinguish a response to a command
            from another.  (It is not incremented for ASYNC
            invocation.)  Note that this variable (field) should be
            declared as byte, but it seems triggering a strange bug on
            some VM.)  */
	private int serial = 0;

	/** The report buffer.  One byte for a report ID and eight
            bytes for the report contents.  This buffer is used both
            for input and output.  <br> The name of this variable is
            too generous and inappropriate.  FIXME.  */
	private byte[] data = new byte[9];

	/** The report received previously.  This is used to
            double-check the received report against a nasty bug.
            FIXME.  */
	private byte[] prevdata = new byte[9];

	/** Invoke a command on this USBIO.  <code>cmd</code> must be
            a valid USBIO command code, optionally bit-wise-or'ed with
            one of the command modifiers defined as static final
            variables of UsbIo class.  <code>arg</code> must be a
            one-byte argument to the command.  <code>arg</code> must
            be specified even if the command takes no arguments.  (The
            argument will be passed to USBIO device and then discarded
            in this case.)  invoke(int, int) returns the response from
            the USBIO device as the result.  (The returned value is
            indeterminable when the command returns no meaningful
            value or when ASYNC modifier is specified.  */
	public int invoke ( final int cmd, final int arg )
	{
	    synchronized ( lock ) {

		/* The RECEIVE pseudo command is special, so we take
		   care of it separately.  */
		if ( cmd == RECEIVE ) {

		    /* Just receive Input Report.  We ignore any error
		       condition in case of REPORT, since this pseudo
		       command is unreliable anyway.  (The data[2]
		       should hold a value left on the last
		       invocation, when the API call failed.)  */
		    theInstance.receiveFromDevice( fd, data );
		    return data[2];
		}

		boolean result;

		/* If this is a synchronous transaction, allocate a
                   new serial number for it.  */
		if ( ( cmd & ASYNC ) == 0 ) {
		    serial++;
		}

		/* Build and send a USBIO command packet in the form
                   of OUTPUT report.  */
		data[0] = 0; /* report ID; always a zero.  */
		data[1] = ( byte )cmd;
		data[2] = ( byte )arg;
		data[3] = ( byte )serial;

		result = theInstance.sendToDevice( fd, data );
		if ( !result ) {
		    throw new UsbIoException( "sendToDevice failed" );
		}

		/* If the request was async, we don't wait for the
                   result.  We return a -1 to indicate there is no
                   returned value.  */
		if ( ( cmd & ASYNC ) != 0 ) {
		    return -1;
		}

		/* Receive a response.  We only need first four byte
		   (i.e., one report ID byte and three payload
		   bytes1), but we need to read nine bytes, because
		   the USB IO device's report format consists of fixed
		   eight bytes.  */
		/* The number of maximum retry count (32) is put just
		   arbitrarily.  FIXME.  */
		byte[] got = null;
		for ( int retry = 0; retry < 32; retry++ ) {

//		    /* It may take several milliseconds before we can
//		       receive the updated response.  Give other
//		       activities some chance to get scheduled, while
//		       we are waiting.  */
//		    try {
//			Thread.sleep( 1 );
//		    } catch ( InterruptedException e ) {
//		    }

		    /* Receive Response.  */
		    result = theInstance.receiveFromDevice( fd, data );
		    if ( !result ) {
			throw new UsbIoException( "receiveFromDevice failed" );
		    }
		
		    /* USB IO firmware always acknowledges the IN
		       transaction on its Interrupt Input pipe
		       (endpoint 1).  I guess that's the reason we
		       receive many delayed INPUT reports through the
		       ReadFile.  (I need a protocol analyzer to
		       verify it...)  Anyway, we have added a serial
		       number on the OUTPUT report, and the device
		       just send it back, so we can make sure the
		       command has been handled.  */
		    if ( data[3] == ( byte )serial ) {

			/* For whatever reason, We sometimes (e.g.,
                           three out of 10,000 experiments) receive a
                           _broken_ input report on some platform
                           (particularly, Windows 98.)  We make sure
                           the received report is stable, by reading
                           repeatedly until getting same content.  */
			if ( NO_DOUBLE_CHECK
			     || arrayEquals( data, prevdata ) ) break;

			/* If the previous data has a same serial
                           number, we have just hit the problem.  */
			if ( VERBOSE_DOUBLE_CHECK
			     && prevdata[3] == serial ) {
			    System.err.println( "xyzzy.hardware.UsbIo:"
						+ " Double Check Failure!" );
			}

			/* Save the data as prevdata for next double check.  */
			System.arraycopy( data, 0, prevdata, 0, data.length );
		    }
		}
	    }

	    /* Verify the received data.  */
	    if ( data[0] != 0 || data[1] != cmd ) {
		throw new UsbIoException( "Invalid response" );
	    }

	    /* Return the received byte as the result.  Remember that
	       Java byte is a signed number and we don't want
	       returning negative number upon successful return.  */
	    return data[2] & 255;
	}
    }

    public static class UsbIoException
        extends RuntimeException
    {
	public UsbIoException ()
	{
	    super();
	}

	public UsbIoException ( final String msg )
	{
	    super( msg );
	}
    }

    /* Morphy USBIO identification.  WARNING: Modifying any member of
       this section may not affect the behaviour of a subclass.
       Platform-dependent implementations are allowed not to refer to
       these variables, since JNI specification provides no handy way
       to do so.  */

    /** The USB Vendor ID of the Morphy USBIO.  It is 0x0BFE.  In this
        version, only devices that have vendor id exactly matching
        with this value are considered as USBIO.  */
    protected static final int targetVendorId = 0x0BFE;

    /** The USB Product ID of the Morphy USBIO.  It is 0x1000.  In
        this version, only devices that have product id matching with
        this value ignoring several bits specified by
        targetProductIdMask are considered as USBIO.  */
    protected static final int targetProductId = 0x1000;

    /** The USB Product ID mask.  It is 0x000F.  See explanation of
        targetProductId for the purpose of targetProductIdMask.  */
    protected static final int targetProductIdMask = 0x000F;

    /** A _callback_ that may be used to identify Morphy USBIO.  This
        method exists just to support development of
        platform-dependent implementation.  Subclasses of UsbIo are
        allowed not to call this method.  */
    protected boolean identifyDevice (
	final short vid, final short pid, final short revision )
    {
	return vid == targetVendorId
	    && ( pid & ~targetProductIdMask ) == targetProductId;
    }

    /* Platform dependent methods.  */

    /** Interact with underlaying platform to find all USBIO devices
        attached to the machine.  findDevices() returns an
        array of handles.  A handle is an opaque object that contains
        all information needed to identify a USB target device on the
        system.  Actual internal structure of a handle is platform
        dependent.  */
    abstract protected Object[] findDevices ();
    
    /** Interact with underlying platform to let the USBIO device
	specified by <code>handle</code> be available to use.
	openDevice(Object) returns an opaque int value to identify
	communication channel to the underlying platform.  It is
	intended to be something like a file descriptor, but the
	actual meaning may vary depending on the platform.  Upon
	success, the returned value must be a positive value
	(including a zero), and a negative (excluding a zero)
	otherwise.  */
    abstract protected int openDevice ( Object handle );

    /** Interact with underlying platform to quit the communication
        channel identified by <code>fd</code>.  */
    abstract protected void closeDevice ( int fd );

    /** Interact with underlying platform to send (output) a report.
        The <code>fd</code> is a communication channel identifier
        returned by openDevice(Object).  <code>report</code> must be
        an array of bytes of length nine.  The array must be filled by
        the caller before calling this method.  The first byte of the
        <code>report</code> is the REPORT ID (always zero for Morphy
        USBIO.)  Other eight bytes are the report content to send.
        The method returns true upon success, false otherwise.<br>

        The call may block until the underlying SET REPORT request has
        been completed, or it may return before completion
        (immediately after scheduling the request.)  The exact
        behaviour can vary on the underlying platform and/or
        implementation class.  */
    abstract protected boolean sendToDevice ( int fd, byte[] report );

    /** Interact with underlying platform to receive (input) a report.
        The <code>fd</code> is a communication channel identifier
        returned by openDevice(Object).  <code>report</code> must be
        an array of bytes of length nine.  The array will be filled by
        the method upon return.  The first byte of the
        <code>report</code> is the REPORT ID (always zero for Morphy
        USBIO.)  Other eight bytes are received report content.  The
        method returns true upon success, false otherwise.<br>

        The call will be block until a report is received from the
        device.  If there were two or more pending reports, the method
        may receive the oldest report, keeping other reports, or it
        may receive the newest one, discarding others.  The exact
        behaviour can vary on the underlying platform and/or
        implementation class.  */
    abstract protected boolean receiveFromDevice ( int fd, byte[] report );

    /* support routines.  */

    /** arrayEquals compares contents equality of two byte arrays.  */
    protected static boolean arrayEquals ( final byte[] a, final byte[] b )
    {
	/* We don't compare length, since we know they are equal.  */
	for ( int i = 0; i < a.length; i++ ) {
	    if ( a[i] != b[i] ) return false;
	}
	return true;
    }
}
