/*
USBIO Support Program
Copyright (C) 2001 by Masahiro Sekiguchi, all rights reserved.

Redistribution and use in source and/or binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote
   products derived from this software.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package xyzzy.hardware;

/** An SPI device driver connected to a USBIO device.  We asseme the
    following wireing: (1) The SPI clock signal is assigned to a bit
    on Port #1, (2) The bi-directional data signal is assigned to a
    bit on Port #0, (3) The SPI select signal is driven by one or more
    bits on Port #0.

    The clock signal is controlled through "read with strobe" and
    "write with strobe" commands of USBIO, so it can be seen both as
    active-low level-triggered or as falling-edge-triggered-out
    raising-edge-triggered-in.  USBIO device works significantly slow,
    so various (leading, trailing, precharge, delay, etc.) time
    constraints are automatically satisfied in most applications,
    although it is impossible to satisfy opposite constraints such as
    maximum operational duration of MPC3404 ADC, which requires a
    controller to complete a command cycle within 1.2ms, while it
    usually takes more than 100ms to complete the cycle with USBIO!

    The SPI select signal needs special consideration.  It is a common
    practice that clock and data signals of two or more SPI target
    chips are tied together with dedicated select signals.  Our SPI
    controller design allows such configuration.  two or more bits in
    Port#0 can be allocated to drive one of SPI select signals.  If
    the number of SPI target chips is small, those select signals can
    be connected to apropriate bits on Port #0.  For example, if you
    have three SPI targets, those select signals can be tied to bits
    0, 1 and 2, directly.  If you have many SPI devices or you need
    extra I/O signals other than SPI communication, you can use an
    external binary decoder to control select signals.  For example,
    you can use a 74HC138 to expands three bits on Port #0 into seven
    select signals to control seven SPI targets.  (You cannot use
    eighth decoded output, since we need a deselection state.  See
    below for details.
    
  */

public class UsbIoSpiDevice
    extends SpiDevice
{
    protected final UsbIo.Device dev;

    protected final int select;
    protected final int deselect;
    protected final int selMask;

    protected final int dataInMask;
    protected final int dataOutMask;

    protected final int WR0_PX;
    protected final int RD0_PX;

    protected static final int WR0 = UsbIo.WRITE0;
    protected static final int WR1 = UsbIo.WRITE1;
    protected static final int RD0 = UsbIo.READ0;

    public UsbIoSpiDevice (
	final UsbIo.Device device,
	final int clock, final int dataIn, final int dataOut, 
	final int select, final int deselect, final int selMask )
    {
	super( device.getLockObject() );
	this.dev = device;

	switch ( clock ) {
	  case 0: WR0_PX = UsbIo.WRITE0_P0; RD0_PX = UsbIo.READ0_P0; break;
	  case 1: WR0_PX = UsbIo.WRITE0_P1; RD0_PX = UsbIo.READ0_P1; break;
	  case 2: WR0_PX = UsbIo.WRITE0_P2; RD0_PX = UsbIo.READ0_P2; break;
	  case 3: WR0_PX = UsbIo.WRITE0_P3; RD0_PX = UsbIo.READ0_P3; break;
	  default:
	    WR0_PX = 0; RD0_PX = 0;
	    exception( "Invalid clock bit number: " + clock );
	}

	if ( dataIn < 0 && dataIn > 7 ) {
	    exception( "Invalid dataIn bit number: " + dataIn );
	}
	dataInMask = 1 << dataIn;
	if ( dataOut < 0 && dataOut > 7 ) {
	    exception( "Invalid dataOut bit number: " + dataOut );
	}
	dataOutMask = 1 << dataOut;

	if ( ( select & ~255 ) != 0 ) {
	    exception( "Invalid select bit pattern: " + select );
	}
	if ( ( deselect & ~255 ) != 0 ) {
	    exception( "Invalid deselect bit pattern: " + deselect );
	}
	if ( ( selMask & ~255 ) != 0 ) {
	    exception( "Invalid selMask bit pattern: " + selMask );
	}
	if ( ( select & ~selMask ) != 0 ) {
	    exception( "Invalid select/selMask combination: "
		       + select + ", " + selMask );
	}
	if ( ( deselect & ~selMask ) != 0 ) {
	    exception( "Invalid deselect/selMask combination: "
		       + deselect + ", " + selMask );
	}
	this.select = select;
	this.deselect = deselect;
	this.selMask = selMask;

	/* Deselect (reset) all SPI chips and put the clock line high.  */
	final int p0 = dev.invoke( RD0, 0 ) & ~selMask | ioMask;
	dev.invoke( WR0_PX, p0 | deselect | dataOutMask );
    }

    /** Abrebiated constractor for bidirectional data line configuration.  */
    public UsbIoSpiDevice (
	final UsbIo.Device device, final int clock, final int data,
	final int select, final int deselect, final int selMask )
    {
	this( device, clock, data, data, select, deselect, selMask );
    }

    /** Abribiated constractor for bidirectional data line and single
        select line configuration (i.e., only one device at all.)  In
        this constractor, the select pin is designated by just one
        number in range 0 to 7.  The pin specified by the number will
        be the active-low select outout.  */
    public UsbIoSpiDevice (
	final UsbIo.Device device,
	final int clock, final int data, final int select )
    {
	this( device, clock, data, data, 0, 1 << select, 1 << select );
    }

    /** The raw I/O Mask word set by the application.  */
    protected int ioMask = 0x00;

    /** Set the I/O mask bits for USBIO port #0.  The I/O mask is
        considered as a series of bits of length 8.  When a bit in the
        I/O mask is set, the corresponding bit in the USBIO port is
        considered to be a readonly.  When we write into a port, bits
        designated as read-only is written as ones.  Otherwise, we
        write a value same as read from the bit.  Note that the i/O
        mask is only provided for port#0; all Port#1 bits are
        effectively considered as output port, since we use read/write
        with strobe commands.  */
    public void setIoMask ( final int mask )
    {
	if ( ( mask & ~255 ) != 0 ) {
	    exception( "Invalid I/O mask: " + mask );
	}
	if ( ( mask & selMask ) != 0 ) {
	    exception( "Overrapping I/O mask and selMask: "
		       + mask + ", " + selMask );
	}
	if ( ( mask & ( dataInMask | dataOutMask ) ) != 0 ) {
	    exception( "Overrapping I/O mask and data: " + mask );
	}
	ioMask = ( mask | dataInMask ) & ~dataOutMask;
    }

    public int getIoMask ()
    {
	return ioMask & dataInMask;
    }

    public void close ()
    {
	synchronized ( getLockObject() ) {

	    /* Make sure to deselect all SPI chips.  */
	    final int p0 = dev.invoke( RD0, 0 ) & ~selMask | ioMask;
	    dev.invoke( WR0_PX, p0  | deselect | dataOutMask );

//	    /* Put the USBIO device in the RESET state.  */
//	    dev.invoke( WR1, 0xFF );
//	    dev.invoke( WR0, 0xFF );

	    /* Close the underlying UsbIo.Device object.  */
	    dev.close();
	}
    }

    /** Issue a command to an SPI chip.  */
    public long execute ( final long outBits, final int outLen,
			  final int inLen )
    {
	synchronized ( getLockObject() ) {

	    /* Save the current states of unused port0 bits.  */
	    final int port0
		= ( dev.invoke( RD0, 0 ) | ioMask ) & ~selMask & ~dataOutMask;

	    /* Send out bits.  (Selection will be done
               automatically/simultaneously.)  */
	    if ( isLsbFirst() ) {
		long m = 1;
		for ( int i = 0; i < outLen; i++ ) {
		    if ( ( outBits & m ) == 0 ) {
			dev.invoke( WR0_PX, port0 | select );
		    } else {
			dev.invoke( WR0_PX, port0 | select | dataOutMask );
		    }
		    m <<= 1;
		}
	    } else {
		long m = 1 << outLen >> 1;
		for ( int i = 0; i < outLen; i++ ) {
		    if ( ( outBits & m ) == 0 ) {
			dev.invoke( WR0_PX, port0 | select );
		    } else {
			dev.invoke( WR0_PX, port0 | select | dataOutMask );
		    }
		    m >>= 1;
		}
	    }

	    /* The following WR0 is to ensure that we can read signal
               from dataIn bit.  However, it is needed only when (1)
               dataIn and dataOut are same, (2) the last output bit
               was a zero, and (3) we are about to read at least one
               bit.  Otherwise, we can skip it.  As our USBIO is slow,
               reducing one command will reduce a greate amount of
               delay (approx 4ms,) so we should implement this
               optimization in a furure.  FIXME.  */
	    dev.invoke( WR0, port0 | select | dataOutMask );

	    /* Receive specified number of bits.  */
	    long inBits = 0;
	    if ( isLsbFirst() ) {
		long m = 1;
		for ( int i = 0; i < inLen; i++ ) {
		    if ( ( dev.invoke( RD0_PX, 0 ) & dataInMask ) != 0 ) {
			inBits |= m;
		    }
		    m <<= 1;
		}
	    } else {
		long m = 1 << inLen >> 1;
		for ( int i = 0; i < inLen; i++ ) {
		    if ( ( dev.invoke( RD0_PX, 0 ) & dataInMask ) != 0 ) {
			inBits |= m;
		    }
		    m >>= 1;
		}
	    }

	    /* Done.  Deselect SPI chips.  */
	    dev.invoke( WR0, port0 | deselect | dataOutMask );

	    /* That's all.  */
	    return inBits;
	}
    }

    /** Sence the dataIn line with chip selection and no clocking.
        This method is intended to be used by SEEPROM clients after
        write operation.  */
    public int poll ()
    {
	synchronized ( getLockObject() ) {

	    /* Save the current states of unused port0 bits.  */
	    final int port0
		= ( dev.invoke( RD0, 0 ) | ioMask ) & ~selMask & ~dataOutMask;

	    /* Select the target SPI chip without clock.  */
	    dev.invoke( WR0, port0 | select | dataOutMask );

	    /* Sence the data line.  */
	    final int d = ( dev.invoke( RD0, 0 ) & dataInMask ) != 0 ? 1 : 0;

	    /* Deselect the chip.  */
	    dev.invoke( WR0, port0 | deselect | dataOutMask );

	    /* Return the polled value.  */
	    return d;
	}
    }
}
