/*
USBIO Support Program
Copyright (C) 2001 by Masahiro Sekiguchi, all rights reserved.

Redistribution and use in source and/or binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote
   products derived from this software.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package xyzzy.hardware.jdirect;

import java.util.Vector;
import com.ms.win32.*;
import com.ms.dll.DllLib;

/** Win32 implementation of UsbIo using J/Direct (Microsoft JavaVM
    only.)  */
public final class UsbIo
    extends xyzzy.hardware.UsbIo
    implements xyzzy.hardware.MachDep
{
    static {
	/* We need to do <em>something</em> that depends on J/Direct
	   upon class initialization, since otherwise, when this code
	   is executed under no Microsoft VM, the MachDepManager may
	   not know this class is <em>not</em> compatible with the
	   current VM.  Loading Kernel32 will be a good choice, since */
	Kernel32.GetVersion();
    }	

    protected Object[] findDevices ()
    {
	/* Find the GUID for HID device class.  I don't understand why
           this value is not a constant, except that I'm writing for
           Windows.  The GUID datatype is just an array of byte of
           length 16, in our wrapper.  */
	final byte[] hidGUID = new byte[16];
	Hid.HidD_GetHidGuid( hidGUID );
	
	/* Get a handle to the set of all HID devices.  The handle
           must *absolutely* be destroyed (freed) later; otherwise the
           memory leaks.  We do our best, using several try blocks,
           but we cannot override Java stupidity.  So if a user hit a
           BREAK, WIndows' system memory will leak.  *sign* */
	final int /*HDEVINFO*/ deviceInfoSet
	    = SetupApi.SetupDiGetClassDevs(
		hidGUID, null, 0,
		( SetupApi.DIGCF_PRESENT | SetupApi.DIGCF_DEVICEINTERFACE ) );
	try {
	    final Vector list = new Vector();

	    /* Scan the set of HID devices for USBIO devices.  */
	    for ( int index = 0;; index++ ) {
		boolean result;

		/* Extract some information about the device at _index_.  */
		final SP_DEVICE_INTERFACE_DATA deviceInterfaceData
		    = new SP_DEVICE_INTERFACE_DATA();
		deviceInterfaceData.cbSize
		    = DllLib.sizeOf( deviceInterfaceData );
		result = SetupApi.SetupDiEnumDeviceInterfaces(
		    deviceInfoSet, null, hidGUID, index, deviceInterfaceData );

		/* The false returned from
		   SetupApi.SetupDiEnumDeviceInterfaces indicates
		   there are no more HID devices attached to the
		   machine.  That is, we have scanned all HID devices.  */
		if ( !result ) break;

		/* Extract the path to the device (at _index_.)  */
		final SP_DEVICE_INTERFACE_DETAIL_DATA deviceInterfaceDetailData
		    = new SP_DEVICE_INTERFACE_DETAIL_DATA();
		/* DDK document says that cbSize field in
		   SP_DEVICE_INTERFACE_DETAIL_DATA must be set to the
		   size of the struct.  Sounds reasonable.  What is
		   not reasonable is, however, the size must be of the
		   template struct as declared in C++ header file come
		   with the DDK, which is *always* different from the
		   actual size of the struct passed to APIs...  That's
		   probably the reason why we have to pass the actual
		   size of the struct to the API as a separate
		   parameter...  This is the second worst
		   brain-damaged API design!  (Analyze your DDK header
		   <setupapi.h> to see why the size must be as
		   follows.)  */
		deviceInterfaceDetailData.cbSize
		    = 4 + DllLib.systemDefaultCharSize;
		result = SetupApi.SetupDiGetDeviceInterfaceDetail(
		    deviceInfoSet,
		    deviceInterfaceData, deviceInterfaceDetailData,
		    DllLib.sizeOf( deviceInterfaceDetailData ), null, null );
		if ( !result ) continue;
		final String devicePath = deviceInterfaceDetailData.devicePath;
		
		/* Now, try to open the path and see if this is the
		   target device.  */
		final int dev = openDevice( devicePath );
		if ( dev >= 0 ) {
		    try {

			/* Extract device attribute.  */
			final HIDD_ATTRIBUTES deviceAttr
			    = new HIDD_ATTRIBUTES();
			deviceAttr.size
			    = DllLib.sizeOf( deviceAttr );
			if ( !Hid.HidD_GetAttributes( dev, deviceAttr ) ) {
			    /* GetAttributes should not fail in normal
			       operation.  I'm not sure why it does.
			       Just ignore this device.  */
			    continue;
			}
			
			/* See if this is a Morphy USBIO.  */
			if ( identifyDevice( deviceAttr.VendorID,
					     deviceAttr.ProductID,
					     deviceAttr.VersionNumber ) ) {
				/* Ah, this device just looks like a
				   USBIO.  Add it to the list.  */
			    list.addElement( devicePath );
			}
			
		    } finally {
			/* We anyway close the underlying file handle
			   for now.  The application must re-issue the
			   Device.open() to access the USBIO device
			   again.  */
			closeDevice( dev );
		    }
		}
	    }

	    /* We found the (possibly empty) list of devices.  */
	    Object[] devList = new Object[list.size()];
	    list.copyInto( devList );
	    return devList;

	} finally {
	    SetupApi.SetupDiDestroyDeviceInfoList( deviceInfoSet );
	}
    }

    protected int openDevice ( final Object handle )
    {
	/* Open the path for the device through as-always CreateFile
	   API to obtain the device handle.  */
	return Kernel32.CreateFile( ( String )handle,
				    win.GENERIC_READ | win.GENERIC_WRITE,
				    win.FILE_SHARE_READ | win.FILE_SHARE_WRITE,
				    null,
				    win.OPEN_EXISTING,
				    0,
				    0);
    }

    protected void closeDevice ( final int fd )
    {
	Kernel32.CloseHandle( fd );
    }

    protected boolean sendToDevice ( final int fd, final byte[] data )
    {
	/* We don't want to allocate an extra array here, but, in
           Java, there are no easy way to reduce the allocation
           overhead of this array.  FIXME.  */
	final int[] count = new int[1];
	final boolean r = Kernel32.WriteFile( fd, data, 9, count, null );
	return r && count[0] == 9;
    }

    protected boolean receiveFromDevice ( final int fd, final byte[] data )
    {
	/* See corresponding comments in sendToDevice on count.
           FIXME.  */
	final int[] count = new int[1];
	final boolean r = Kernel32.ReadFile( fd, data, 9, count, null );
	return r && count[0] == 9;
    }
}
