/*
USBIO Support Program
Copyright (C) 2001 by Masahiro Sekiguchi, all rights reserved.

Redistribution and use in source and/or binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote
   products derived from this software.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

/* The following four headers are apparently needed to access to USB
   HID devices in FreeBSD.  We *don't* need <libusb.h>, since this
   program doesn't parse HID descriptors/reports.  */
#include <sys/types.h>
#include <sys/ioctl.h>
#include <dev/usb/usb.h>
#include <dev/usb/usbhid.h>

/* Then comes JNI staff.  */
#include <jni.h>
#include "../xyzzy_hardware_jni_UsbIo.h"

/* Forward declarations, as always.  */
static jboolean waitReady ( JNIEnv *, int, jboolean );
static void throwNew ( JNIEnv *, char const *, char const * );

/* Various implicit constants (mostly strings.)  */

/* The device path prefix.  USB device files (inodes) are assumed to
   be located in the following directory.  Note that DEVPREFIX is
   considered as a literal prefix, so the '/' at the end is mandatory.  */
#define DEVPREFIX	"/dev/"

/* The USB bus (controller) device name.  Only filenames starting with
   this prefix followed by a digit are assumed to be USB bus devices.
   Anything else will be ignored.  (In this version, we cannot handle
   two or more different prefixes.  Fortunately, FreeBSD defines the
   USB bus device as abstract device that is independent from
   underlying controller chip (i.e., UHCI v.s. OHCI.)  So it should be
   sufficient to handle only one name.)  */
#define USBDEVICE	"usb"

/* Find all USBIO devices on the system.  */
JNIEXPORT jobjectArray JNICALL
Java_xyzzy_hardware_jni_UsbIo_findDevices
    ( JNIEnv * env, jobject usbio )
{
    int i;
    int count;
    int unit;
    char * busname = alloca( strlen( DEVPREFIX USBDEVICE ) + 2 );
    int fd;
    int addr;
    struct usb_device_info devinfo;
    size_t len;
    jbyteArray ba;
    struct devList {
	jbyteArray ba;
	struct devList const * next;
    };
    struct devList const * list = NULL;
    struct devList * dlp;
    jclass aClass;
    jobjectArray theList;
    jmethodID identifyDevice;

    /* Prepare for invoking UsbIo.identifyDevice(short,short,short)
       method.  */
    aClass = (*env)->GetObjectClass( env, usbio );
    identifyDevice
	= (*env)->GetMethodID( env, aClass, "identifyDevice", "(SSS)Z" );
    aClass = NULL;

    /* The variable count tracks how many USBIO devices are found.  */
    count = 0;

    /* Firstly, we list up all USB bus devices.  We only examine up to
       ten devices, since the specification (of what? FIXME :-) says
       only one digit can follow the USBDEVICE prefix.  (However, in
       the future the USB bus device _digit_ may be extended to cover
       'a' to 'f' or even more, just like what happened on pseudo tty
       devices...  (Remember that the tenth pseudo tty device is
       called "/dev/ttypa" and the last character of this name is
       called "digit.")  FIXME.)  */
    for ( unit = 0; unit < 10; unit++ ) {
	sprintf( busname, "%s%d", DEVPREFIX USBDEVICE, unit );

	fd = open( busname, O_RDONLY );
	if ( fd < 0 ) continue;
	
	/* We've successfully opened a USB bus device.  Next, we
           enumerate all USB target devices, looking for USBIO
           devices.  */
	for ( addr = 0; addr < USB_MAX_DEVICES; addr++ ) {

	    /* Ask for the USB device information.  */
	    memset( &devinfo, 0, sizeof ( devinfo ) );
	    devinfo.addr = addr;
	    if ( ioctl( fd, USB_DEVICEINFO, &devinfo ) < 0 ) {
		/* Skip missing devices.  */
		continue;
	    }

	    /* Ask our Java pier to identify USBIO device.  */
	    if ( (*env)->CallBooleanMethod( env,
					    usbio,
					    identifyDevice,
					    ( jshort )devinfo.vendorNo,
					    ( jshort )devinfo.productNo,
					    ( jshort )devinfo.releaseNo )
		 == JNI_FALSE ) {
		continue;
	    }

	    /* This device looks like a USBIO.  Record the device name
               as a C-style (narrow) string that is represented in
               Java as a byte array.  XXX: In this version, we only
               use the first device driver (devnames[0]) attached to
               this target.  This may cause some trouble in the
               future.  FIXME.  */
	    len = strlen( devinfo.devnames[0] ) + 1;
	    ba = (*env)->NewByteArray( env, len );
	    (*env)->SetByteArrayRegion( env, ba, 0, len,
					devinfo.devnames[0] );
	    
	    /* Put the array in the linked list of Device Paths.  */
	    dlp = alloca( sizeof( struct devList ) );
	    dlp->ba = ba;
	    dlp->next = list;
	    list = dlp;

	    /* We found another USBIO device.  Don't forget to count up.  */
	    count++;
	}

	/* We have done on this USB bus.  Close the bus device.  */
	close( fd );
    }

    /* We no more need identify method.  */
    identifyDevice = NULL;

    /* We should have found all USBIO devices at this point.
       Corresponding handle objects (Java byte arrays) are kept with a
       linked list.  We allocate another Java array to store those
       handles and return it.  */
    aClass = (*env)->FindClass( env, "java/lang/Object" );
    theList = (*env)->NewObjectArray( env, count, aClass, NULL );
    for ( i = count - 1; i >= 0; --i ) {
	(*env)->SetObjectArrayElement( env, theList, i, list->ba );
	list = list->next;
    }
    return theList;
}

JNIEXPORT jint JNICALL
Java_xyzzy_hardware_jni_UsbIo_openDevice
    ( JNIEnv * env, jobject usbio, jobject handle )
{
    jbyteArray array = ( jbyteArray )handle;
    char * path;
    size_t plen;	/* Length of prefix excluding terminating NUL. */
    size_t hlen;	/* Length of device name including terminating NUL. */

    /* Extract the C-style string for the device name, prepending the
       device file prefix.  (Note that the array is a C-style
       NUL-terminated string represented as a Java array of exact
       length.  So, we need not re-evaluate the string length.  */
    plen = strlen( DEVPREFIX );
    hlen = (*env)->GetArrayLength( env, array );
    path = alloca( plen + hlen );
    strcpy( path, DEVPREFIX );
    (*env)->GetByteArrayRegion( env, array, 0, hlen, path + plen );

    /* Now, open the specified path and return it.  As FreeBSD JavaVM
       only support green threads (as of early 2001), any blocking
       system call issued in JNI native code stops entire JVM.  I set
       O_NOBLOCK here, although it doesn't work well on FreeBSD.  In
       this version, we use select() to switch context to avoid
       blocking.  See waitReady() below.  */
    return open( path, O_RDWR /* | O_NONBLOCK */ );
}

JNIEXPORT void JNICALL 
Java_xyzzy_hardware_jni_UsbIo_closeDevice
    ( JNIEnv * env, jobject usbio, jint fd )
{
    close( fd );
}

/* Send a command to USBIO device.  */
JNIEXPORT jboolean JNICALL 
Java_xyzzy_hardware_jni_UsbIo_sendToDevice
    ( JNIEnv * env, jobject usbio, jint fd, jbyteArray data )
{
    jbyte buffer[8];

    if ( (*env)->GetArrayLength( env, data ) != 9 ) {
	throwNew( env, "java/lang/IllegalArgumentException",
		  "invalid REPORT data size" );
	return JNI_FALSE;
    }

    /* As of FreeBSD 4.3, write() to uhid device is totally
       synchronous; i.e., the call to the write() always blocks,
       waiting for the completion of the underlying SETUP trasfer.
       So, the following waitReady() is useless.  It means we cannot
       avoid stopping entire JavaVM during sendToDevice, since there
       is no way to return from the write() system call before
       receiving ACK from the device.  */

    if ( !waitReady( env, fd, JNI_FALSE ) ) return JNI_FALSE;

    /* Write the command (report content) portion in the data.  */
    (*env)->GetByteArrayRegion( env, data, 1, sizeof ( buffer ), buffer );
    return write( fd, buffer, sizeof ( buffer ) ) == sizeof ( buffer);
}

JNIEXPORT jboolean JNICALL 
Java_xyzzy_hardware_jni_UsbIo_receiveFromDevice
    ( JNIEnv * env, jobject usbio, jint fd, jbyteArray data )
{
    /* Note the trick of greed: it is the 64 byte array in mind, with
       an extra byte before them.  The extra byte is used to
       supplement the REPORT ID that is (for whatever reason) removed
       by the FreeBSD kernel.  Also note that the size 64 (without
       counting the extra byte) is chosen to be an integral multiple
       of 8 (the size of the response (REPORT content)) and be smaller
       than the _chunk_buffer_ (128 bytes) in the uhid driver.  */
    jbyte greed[65];
    jbyte * p;
    int n;

    if ( (*env)->GetArrayLength( env, data ) != 9 ) {
	throwNew( env, "java/lang/IllegalArgumentException",
		  "invalid REPORT data size" );
	return JNI_FALSE;
    }

    if ( !waitReady( env, fd, JNI_TRUE ) ) return JNI_FALSE;

    /* We use a large (but not too large) buffer to receive responses
       (input reports) from the kernel, and use the last one as the
       newest input.  This way, we can effectively discard aged
       reports queued in the kernel buffer.  (Should we peek into the
       buffer, e.g. using select(), to pull queued reports more
       aggressively?  FIXME.)  */
    n = read( fd, greed + 1, sizeof ( greed ) - 1 );
    if ( n < 8 || n % 8 != 0 ) return JNI_FALSE;

    /* Copy the _last_ eight bytes of the read data to the caller
       supplied byte array, prepending a 0x00 as a REPORT ID (that is
       always a zero in USBIO.)  */
    p = greed + n - 8;
    p[0] = 0;
    (*env)->SetByteArrayRegion( env, data, 0, 9, p );

    return JNI_TRUE;
}

/* Wait for the specified file descriptor is ready for read/write.
   While waiting, we issue java.lang.Thread.sleep(long,int) to give
   other (Java) threads some chance of getting scheduled.  */
static jboolean
waitReady ( JNIEnv * env, int fd, jboolean toRead )
{
    int r;
    jclass threadClass = NULL;
    jmethodID sleep = NULL;
    fd_set * rfds, * wfds;
    fd_set fds;
    static struct timeval zero = {0};

    /* Initialize local stuff.  Zeroing the fd_set is somewhat
       expensive task, so we want to do it only once.  */
    FD_ZERO( &fds );
    if ( toRead ) {
        rfds = &fds;
	wfds = NULL;
    } else {
        rfds = NULL;
	wfds = &fds;
    }

    for (;;) {
        /* See if the descriptor is ready.  */
	FD_SET( fd, &fds );
	r = select( fd + 1, rfds, wfds, NULL, &zero );
	if ( r < 0 ) return JNI_FALSE;
	if ( r > 0 ) return JNI_TRUE;

	/* If not, switch the context.  But before it, we need to
           initialize some JNI stuff, if this is the first time for
           waiting.  Any failure on these JNI calls means Error, so if
           it happens, just return to the Java caller to throw them.  */
	if ( threadClass == NULL ) {
	    threadClass = (*env)->FindClass( env, "java/lang/Thread" );
	    if ( threadClass == NULL ) return JNI_FALSE;
	    sleep = (*env)->GetStaticMethodID( env, threadClass,
					       "sleep", "(JI)V" );
	    if ( sleep == NULL ) return JNI_FALSE;
	}

	/* Sleep for 200us.  The value 200us is chosen so that it is
           enough short relative to the USB frame interval (1ms) and
           that it is long enough relative to the CPU speed (to reduce
           the context switch overhead.)  */
	(*env)->CallStaticVoidMethod( env, threadClass, sleep,
				      ( jlong )0, ( jint )200000 );

	/* If any errors/exceptions are thrown when we were sleep'ing
           (most likely an InterruptedException caused by an
           interrupt() call by another thread), we return immediately,
           indicating a failure.  */
	if ( (*env)->ExceptionOccurred( env ) != NULL ) return JNI_FALSE;
    }
}

/* Throws an Error/Exception, creating a new instance.  This function
   is similar to the JNI's ThrowNew API, except that the Throwable to
   be thrown is specified by its name (using a C string but jclass
   object) and that the pending error/exception (if one exists) will
   be cleared before throwing.  This function is intended to be called
   infrequently, so it is generally better to delay finding the class
   until the function is actually called.  */
static void 
throwNew ( JNIEnv * env, char const * class, char const * msg )
{
    jclass aClass;

    /* Clear the pending Throwable first, if one exists, since we want
       to throw another.  (JNI keeps previous one unless we explicitly
       clear it.  */
    (*env)->ExceptionClear( env );

    /* Find the class to create a new object and throw.  */
    aClass = (*env)->FindClass( env, class );

    /* If the above call was failed, the ClassNotFoundException should
       have been thrown.  We just accept it.  */
    if ( aClass == NULL ) return;

    /* Create and throw.  */
    (*env)->ThrowNew( env, aClass, msg );
}

