/*
USBIO Support Program
Copyright (C) 2001 by Masahiro Sekiguchi, all rights reserved.

Redistribution and use in source and/or binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote
   products derived from this software.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#include <stddef.h>
#include <string.h>

#include <windows.h>
#include <setupapi.h>
#include <hidsdi.h>

#include <jni.h>
#include "../xyzzy_hardware_jni_UsbIo.h"

/* findDevices follows the corresponding method in J/Direct version
   (xyzzy.hardware.win32jd.UsbIo.findDevices() almost exactly.  Major
   differences are as follows:

   - Java (J/Direct) version uses try-finally blocks extensively, to
     make sure that system resources are properly released when an
     Error or a RuntimeException is thrown by the VM.  C (JNI) version
     doesn't use finally (since C has no such construct :-), and the
     system resources are just releases when they becomes useless.

   Some other (minor) differences include:

   - Java uses a Vector to list devices, but C uses a linked list
     whose link cell are allocated by alloca().

   - C version follows the DDK-document recommended sequence (query,
     allocate, extract) to call SetupDiGetDeviceInterfaceDetail(),
     since, in C, we can cast a raw memory block into a struct.

*/

JNIEXPORT jobjectArray JNICALL
Java_xyzzy_hardware_jni_UsbIo_findDevices
    ( JNIEnv * env, jobject usbio )
{
    /* We are writing in C, so we need to declare all local variable
       here.  */

    int i;
    int count;

    GUID hidGuid;
    HDEVINFO dis;
    BOOL r;
    DWORD size;

    SP_DEVICE_INTERFACE_DATA sdid;
    PSP_DEVICE_INTERFACE_DETAIL_DATA_A pdidd = NULL; /* to gag gcc.  */
    size_t didd_size = 0;

    HANDLE dh;
    HIDD_ATTRIBUTES attr;

    size_t len;
    jbyteArray ba;
    jbyte * barray;

    struct devList {
	jbyteArray ba;
	struct devList const * next;
    };
    struct devList const * list = NULL;
    struct devList * dlp;

    jclass objectClass;
    jobjectArray theList;

    jclass usbioClass;
    jmethodID identifyDevice;

    /* Prepare for invoking UsbIo.identifyDevice(short,short,short)
       method.  */
    usbioClass = (*env)->GetObjectClass( env, usbio );
    identifyDevice
	= (*env)->GetMethodID( env, usbioClass, "identifyDevice", "(SSS)Z" );

    /* Find the HID class GUID.  I'm not sure why this is not a
       compile-time constant, but te DDK document says we have to call
       the following function at runtime.  */
    HidD_GetHidGuid( &hidGuid );

    /* Get a Device Info Set for HID devices.  */
    dis = SetupDiGetClassDevsA(
	&hidGuid, NULL, 0, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE );

    /* Enumerate all devices in the Device Info Set, looking for USBIO
       devices.  */
    count = 0;
    for ( i = 0;; i++ ) {

	/* Extract the Device Context Structure for the Device
           Interface Element of the i'th device in the Device Info Set
           into sdid.  (Please don't ask me meanings of these terms; I
           don't know.)  */
	memset( &sdid, 0, sizeof( sdid ) );
	sdid.cbSize = sizeof( SP_DEVICE_INTERFACE_DATA );
	r = SetupDiEnumDeviceInterfaces( dis, NULL, &hidGuid, i, &sdid );

	/* If the above call returns with a FALSE, that means the
           Device Info Set (dis) contains no more devices.  I.e., we
           have completed the enumeration.  */
	if ( !r ) break;

	/* Then, extract the Device Path to the i'th device, in three
           steps (as suggested in DDK document): Find the size of
           required memory, allocate the memory, and receive
           information.  We allocate this dynamic memory in the JVM
           native stack (by alloca()) to avoid the potentially
           dangerous use of system heap in this context.  */

	/* Step one: Find the size of required memory.  Note that the
           call to SetupDiGetDeviceInterfaceDetailA() must fail with
           ERROR_INSUFFICIENT_BUFFER at this step, since the
           DeviceInterfaceDetailDeta parameter is set to NULL.
           Otherwise, something very strange should be happening.  */
	size = 0;
	r = SetupDiGetDeviceInterfaceDetailA(
	    dis, &sdid, NULL, 0, &size, NULL );
	if ( r || GetLastError() != ERROR_INSUFFICIENT_BUFFER ) continue;

	/* Step two: allocate the memory (if needed.)  */
	if ( size > didd_size ) {
	    didd_size = ( size + 15 ) & ~15;	/* in 16 byte granules.  */
	    pdidd = alloca( didd_size );
	}
	memset( pdidd, 0, didd_size );
	pdidd->cbSize = sizeof( SP_DEVICE_INTERFACE_DETAIL_DATA_A );

	/* Step three: get detailed information into the variable.  */
	r = SetupDiGetDeviceInterfaceDetailA(
	    dis, &sdid, pdidd, size, NULL, NULL );
	if ( !r ) continue;

	/* Open the device path.  The J/Direct version calls
           openDevice(Object) here, but we call Win32 API directly.
           To use openDevice, we need to construct a handle object,
           but it is somewhat expensive in JNI.  So, I want to delay
           the construction until the Device Path is known to be of
           USBIO.  */
	dh = CreateFileA( ( LPCSTR )pdidd->DevicePath,
			  GENERIC_READ | GENERIC_WRITE,
			  FILE_SHARE_READ | FILE_SHARE_WRITE,
			  NULL,
			  OPEN_EXISTING,
			  FILE_ATTRIBUTE_NORMAL,
			  NULL ); 

	/* We should not fail, since the path is what is just passed
           by the system.  If it fails, we simply ignore the case.
           (Or, should we consider the case as fatal and abort the
           entire operation?  FIXME.)  */
	if ( dh == INVALID_HANDLE_VALUE ) continue;

	/* Ask Vendor, Product, and Revision IDs to know whether the
           device is USBIO.  */
	memset( &attr, 0, sizeof ( HIDD_ATTRIBUTES ) );
	attr.Size = sizeof( HIDD_ATTRIBUTES );
	HidD_GetAttributes( dh, &attr );

	/* We no more need device handle.  Close it now to avoid the
           handle to be kept opened forever.  */
	CloseHandle( dh );

	/* Test for the IDs.  */
	if ( (*env)->CallBooleanMethod( env,
					  usbio,
					  identifyDevice,
					  ( jshort )attr.VendorID,
					  ( jshort )attr.ProductID,
					  ( jshort )attr.VersionNumber )
	     == JNI_FALSE ) {
	    continue;
	}
	
	/* This device looks like a USBIO.  Record the Device Path as
           C-style (narrow) string that is represented in Java as a
           byte array.  */
	len = strlen( pdidd->DevicePath );
	ba = (*env)->NewByteArray( env, ( jsize )( len + 1 ) );
	barray = (*env)->GetByteArrayElements( env, ba, NULL );
	memcpy( barray, pdidd->DevicePath, len + 1 );
	(*env)->ReleaseByteArrayElements( env, ba, barray, 0 );

	/* Put the array in the linked list of Device Paths.  */
	dlp = alloca( sizeof( struct devList ) );
	dlp->ba = ba;
	dlp->next = list;
	list = dlp;

	/* We found another USBIO device.  Don't forget to count up.  */
	count++;
    }

    /* We no more need the HID Device Info Set.  Destroy it
       immediately.  It can fail, but we can do nothing when it
       failed.  So, we don't test the return value.  */
    SetupDiDestroyDeviceInfoList( dis );

    /* We should have found all USBIO devices.  Corresponding handle
       objects to be passed to the Java portion are stored in the
       linked list (list).  We now allocate an array and put handles in
       it.  That's what findDevice() should return.  */
    objectClass = (*env)->FindClass( env, "java/lang/Object" );
    theList = (*env)->NewObjectArray( env, count, objectClass, NULL );
    for ( i = count - 1; i >= 0; --i ) {
	(*env)->SetObjectArrayElement( env, theList, i, list->ba );
	list = list->next;
    }
    return theList;
}

JNIEXPORT jint JNICALL 
Java_xyzzy_hardware_jni_UsbIo_openDevice
    ( JNIEnv * env, jobject usbio, jobject handle )
{
    jbyteArray array = ( jbyteArray )handle;
    jbyte * path = (*env)->GetByteArrayElements( env, array, NULL );
    HANDLE fd = CreateFileA( ( LPCSTR )path,
			     GENERIC_READ | GENERIC_WRITE,
			     FILE_SHARE_READ | FILE_SHARE_WRITE,
			     NULL,
			     OPEN_EXISTING,
			     FILE_ATTRIBUTE_NORMAL,
			     NULL );
    (*env)->ReleaseByteArrayElements( env, array, path, JNI_ABORT );
    return ( jint )fd;
}

JNIEXPORT void JNICALL 
Java_xyzzy_hardware_jni_UsbIo_closeDevice
    ( JNIEnv * env, jobject usbio, jint fd )
{
    CloseHandle( ( HANDLE )fd );
}

JNIEXPORT jboolean JNICALL 
Java_xyzzy_hardware_jni_UsbIo_sendToDevice
    ( JNIEnv * env, jobject usbio, jint fd, jbyteArray data )
{
    jbyte * buffer = (*env)->GetByteArrayElements( env, data, NULL );
    DWORD count = 0;
    BOOL r = WriteFile( ( HANDLE )fd, buffer, 9, &count, NULL );
    (*env)->ReleaseByteArrayElements( env, data, buffer, JNI_ABORT );
    return ( r && count == 9 );
}

JNIEXPORT jboolean JNICALL 
Java_xyzzy_hardware_jni_UsbIo_receiveFromDevice
    ( JNIEnv * env, jobject usbio, jint fd, jbyteArray data )
{
    jbyte * buffer = ( *env )->GetByteArrayElements( env, data, NULL );
    DWORD count = 0;
    BOOL r = ReadFile( ( HANDLE )fd, buffer, 9, &count, NULL );
    (*env)->ReleaseByteArrayElements( env, data, buffer, 0 );
    return ( r && count == 9 );
}

